
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#include "modulemd-enums.h"
#include "modulemd.h"
#include "modulemd-build-config.h"
#include "modulemd-buildopts.h"
#include "modulemd-component.h"
#include "modulemd-component-module.h"
#include "modulemd-component-rpm.h"
#include "modulemd-compression.h"
#include "modulemd-defaults.h"
#include "modulemd-defaults-v1.h"
#include "modulemd-dependencies.h"
#include "modulemd-deprecated.h"
#include "modulemd-errors.h"
#include "modulemd-module.h"
#include "modulemd-module-index.h"
#include "modulemd-module-index-merger.h"
#include "modulemd-module-stream.h"
#include "modulemd-module-stream-v1.h"
#include "modulemd-module-stream-v2.h"
#include "modulemd-packager-v3.h"
#include "modulemd-profile.h"
#include "modulemd-rpm-map-entry.h"
#include "modulemd-service-level.h"
#include "modulemd-subdocument-info.h"
#include "modulemd-translation.h"
#include "modulemd-translation-entry.h"
#include "modulemd-obsoletes.h"

#define C_ENUM(v) ((gint) v)
#define C_FLAGS(v) ((guint) v)

/* enumerations from "modulemd-compression.h" */

GType
modulemd_compression_type_enum_get_type (void)
{
static gsize gtype_id = 0;
static const GEnumValue values[] = {
    { C_ENUM(MODULEMD_COMPRESSION_TYPE_DETECTION_FAILED), "MODULEMD_COMPRESSION_TYPE_DETECTION_FAILED", "detection-failed" },
    { C_ENUM(MODULEMD_COMPRESSION_TYPE_UNKNOWN_COMPRESSION), "MODULEMD_COMPRESSION_TYPE_UNKNOWN_COMPRESSION", "unknown-compression" },
    { C_ENUM(MODULEMD_COMPRESSION_TYPE_NO_COMPRESSION), "MODULEMD_COMPRESSION_TYPE_NO_COMPRESSION", "no-compression" },
    { C_ENUM(MODULEMD_COMPRESSION_TYPE_GZ_COMPRESSION), "MODULEMD_COMPRESSION_TYPE_GZ_COMPRESSION", "gz-compression" },
    { C_ENUM(MODULEMD_COMPRESSION_TYPE_BZ2_COMPRESSION), "MODULEMD_COMPRESSION_TYPE_BZ2_COMPRESSION", "bz2-compression" },
    { C_ENUM(MODULEMD_COMPRESSION_TYPE_XZ_COMPRESSION), "MODULEMD_COMPRESSION_TYPE_XZ_COMPRESSION", "xz-compression" },
    { C_ENUM(MODULEMD_COMPRESSION_TYPE_ZCK_COMPRESSION), "MODULEMD_COMPRESSION_TYPE_ZCK_COMPRESSION", "zck-compression" },
    { C_ENUM(MODULEMD_COMPRESSION_TYPE_SENTINEL), "MODULEMD_COMPRESSION_TYPE_SENTINEL", "sentinel" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_enum_register_static (g_intern_static_string ("ModulemdCompressionTypeEnum"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

/* enumerations from "modulemd-defaults.h" */

GType
modulemd_defaults_version_enum_get_type (void)
{
static gsize gtype_id = 0;
static const GEnumValue values[] = {
    { C_ENUM(MD_DEFAULTS_VERSION_ERROR), "MD_DEFAULTS_VERSION_ERROR", "error" },
    { C_ENUM(MD_DEFAULTS_VERSION_UNSET), "MD_DEFAULTS_VERSION_UNSET", "unset" },
    { C_ENUM(MD_DEFAULTS_VERSION_ONE), "MD_DEFAULTS_VERSION_ONE", "one" },
    { C_ENUM(MD_DEFAULTS_VERSION_LATEST), "MD_DEFAULTS_VERSION_LATEST", "latest" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_enum_register_static (g_intern_static_string ("ModulemdDefaultsVersionEnum"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

/* enumerations from "modulemd-errors.h" */

GType
modulemd_error_get_type (void)
{
static gsize gtype_id = 0;
static const GEnumValue values[] = {
    { C_ENUM(MMD_ERROR_UPGRADE), "MMD_ERROR_UPGRADE", "upgrade" },
    { C_ENUM(MMD_ERROR_VALIDATE), "MMD_ERROR_VALIDATE", "validate" },
    { C_ENUM(MMD_ERROR_FILE_ACCESS), "MMD_ERROR_FILE_ACCESS", "file-access" },
    { C_ENUM(MMD_ERROR_NO_MATCHES), "MMD_ERROR_NO_MATCHES", "no-matches" },
    { C_ENUM(MMD_ERROR_TOO_MANY_MATCHES), "MMD_ERROR_TOO_MANY_MATCHES", "too-many-matches" },
    { C_ENUM(MMD_ERROR_MAGIC), "MMD_ERROR_MAGIC", "magic" },
    { C_ENUM(MMD_ERROR_NOT_IMPLEMENTED), "MMD_ERROR_NOT_IMPLEMENTED", "not-implemented" },
    { C_ENUM(MMD_ERROR_MISSING_REQUIRED), "MMD_ERROR_MISSING_REQUIRED", "missing-required" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_enum_register_static (g_intern_static_string ("ModulemdError"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

GType
modulemd_error_enum_get_type (void)
{
static gsize gtype_id = 0;
static const GEnumValue values[] = {
    { C_ENUM(MODULEMD_ERROR_UPGRADE), "MODULEMD_ERROR_UPGRADE", "upgrade" },
    { C_ENUM(MODULEMD_ERROR_VALIDATE), "MODULEMD_ERROR_VALIDATE", "validate" },
    { C_ENUM(MODULEMD_ERROR_FILE_ACCESS), "MODULEMD_ERROR_FILE_ACCESS", "file-access" },
    { C_ENUM(MODULEMD_ERROR_NO_MATCHES), "MODULEMD_ERROR_NO_MATCHES", "no-matches" },
    { C_ENUM(MODULEMD_ERROR_TOO_MANY_MATCHES), "MODULEMD_ERROR_TOO_MANY_MATCHES", "too-many-matches" },
    { C_ENUM(MODULEMD_ERROR_MAGIC), "MODULEMD_ERROR_MAGIC", "magic" },
    { C_ENUM(MODULEMD_ERROR_NOT_IMPLEMENTED), "MODULEMD_ERROR_NOT_IMPLEMENTED", "not-implemented" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_enum_register_static (g_intern_static_string ("ModulemdErrorEnum"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

GType
modulemd_yaml_error_get_type (void)
{
static gsize gtype_id = 0;
static const GEnumValue values[] = {
    { C_ENUM(MMD_YAML_ERROR_OPEN), "MMD_YAML_ERROR_OPEN", "open" },
    { C_ENUM(MMD_YAML_ERROR_PROGRAMMING), "MMD_YAML_ERROR_PROGRAMMING", "programming" },
    { C_ENUM(MMD_YAML_ERROR_UNPARSEABLE), "MMD_YAML_ERROR_UNPARSEABLE", "unparseable" },
    { C_ENUM(MMD_YAML_ERROR_PARSE), "MMD_YAML_ERROR_PARSE", "parse" },
    { C_ENUM(MMD_YAML_ERROR_EMIT), "MMD_YAML_ERROR_EMIT", "emit" },
    { C_ENUM(MMD_YAML_ERROR_MISSING_REQUIRED), "MMD_YAML_ERROR_MISSING_REQUIRED", "missing-required" },
    { C_ENUM(MMD_YAML_ERROR_EVENT_INIT), "MMD_YAML_ERROR_EVENT_INIT", "event-init" },
    { C_ENUM(MMD_YAML_ERROR_INCONSISTENT), "MMD_YAML_ERROR_INCONSISTENT", "inconsistent" },
    { C_ENUM(MMD_YAML_ERROR_UNKNOWN_ATTR), "MMD_YAML_ERROR_UNKNOWN_ATTR", "unknown-attr" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_enum_register_static (g_intern_static_string ("ModulemdYamlError"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

GType
modulemd_yaml_error_enum_get_type (void)
{
static gsize gtype_id = 0;
static const GEnumValue values[] = {
    { C_ENUM(MODULEMD_YAML_ERROR_OPEN), "MODULEMD_YAML_ERROR_OPEN", "open" },
    { C_ENUM(MODULEMD_YAML_ERROR_PROGRAMMING), "MODULEMD_YAML_ERROR_PROGRAMMING", "programming" },
    { C_ENUM(MODULEMD_YAML_ERROR_UNPARSEABLE), "MODULEMD_YAML_ERROR_UNPARSEABLE", "unparseable" },
    { C_ENUM(MODULEMD_YAML_ERROR_PARSE), "MODULEMD_YAML_ERROR_PARSE", "parse" },
    { C_ENUM(MODULEMD_YAML_ERROR_EMIT), "MODULEMD_YAML_ERROR_EMIT", "emit" },
    { C_ENUM(MODULEMD_YAML_ERROR_MISSING_REQUIRED), "MODULEMD_YAML_ERROR_MISSING_REQUIRED", "missing-required" },
    { C_ENUM(MODULEMD_YAML_ERROR_EVENT_INIT), "MODULEMD_YAML_ERROR_EVENT_INIT", "event-init" },
    { C_ENUM(MODULEMD_YAML_ERROR_INCONSISTENT), "MODULEMD_YAML_ERROR_INCONSISTENT", "inconsistent" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_enum_register_static (g_intern_static_string ("ModulemdYamlErrorEnum"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

/* enumerations from "modulemd-module-stream.h" */

GType
modulemd_module_stream_version_enum_get_type (void)
{
static gsize gtype_id = 0;
static const GEnumValue values[] = {
    { C_ENUM(MD_MODULESTREAM_VERSION_ERROR), "MD_MODULESTREAM_VERSION_ERROR", "error" },
    { C_ENUM(MD_MODULESTREAM_VERSION_UNSET), "MD_MODULESTREAM_VERSION_UNSET", "unset" },
    { C_ENUM(MD_MODULESTREAM_VERSION_ONE), "MD_MODULESTREAM_VERSION_ONE", "one" },
    { C_ENUM(MD_MODULESTREAM_VERSION_TWO), "MD_MODULESTREAM_VERSION_TWO", "two" },
    { C_ENUM(MD_MODULESTREAM_VERSION_LATEST), "MD_MODULESTREAM_VERSION_LATEST", "latest" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_enum_register_static (g_intern_static_string ("ModulemdModuleStreamVersionEnum"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

/* enumerations from "modulemd-obsoletes.h" */

GType
modulemd_obsoletes_version_enum_get_type (void)
{
static gsize gtype_id = 0;
static const GEnumValue values[] = {
    { C_ENUM(MD_OBSOLETES_VERSION_ERROR), "MD_OBSOLETES_VERSION_ERROR", "error" },
    { C_ENUM(MD_OBSOLETES_VERSION_UNSET), "MD_OBSOLETES_VERSION_UNSET", "unset" },
    { C_ENUM(MD_OBSOLETES_VERSION_ONE), "MD_OBSOLETES_VERSION_ONE", "one" },
    { C_ENUM(MD_OBSOLETES_VERSION_LATEST), "MD_OBSOLETES_VERSION_LATEST", "latest" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_enum_register_static (g_intern_static_string ("ModulemdObsoletesVersionEnum"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

/* enumerations from "modulemd-packager-v3.h" */

GType
modulemd_packager_version_enum_get_type (void)
{
static gsize gtype_id = 0;
static const GEnumValue values[] = {
    { C_ENUM(MD_PACKAGER_VERSION_ERROR), "MD_PACKAGER_VERSION_ERROR", "error" },
    { C_ENUM(MD_PACKAGER_VERSION_UNSET), "MD_PACKAGER_VERSION_UNSET", "unset" },
    { C_ENUM(MD_PACKAGER_VERSION_TWO), "MD_PACKAGER_VERSION_TWO", "two" },
    { C_ENUM(MD_PACKAGER_VERSION_THREE), "MD_PACKAGER_VERSION_THREE", "three" },
    { C_ENUM(MD_PACKAGER_VERSION_LATEST), "MD_PACKAGER_VERSION_LATEST", "latest" },
{ 0, NULL, NULL }
        };
        if (g_once_init_enter (&gtype_id)) {
            GType new_type = g_enum_register_static (g_intern_static_string ("ModulemdPackagerVersionEnum"), values);
            g_once_init_leave (&gtype_id, new_type);
        }
        return (GType) gtype_id;
        }

/* Generated data ends here */

